%% Copyright (C) 1994,1995, Ellemtel Telecommunications Systems Laboratories
%%-----------------------------------------------------------------------
%% Created:  94-11-18 by Tobbe (tobbe@erix.ericsson.se)
%%
%% Function: Tokenizer for ANSI C
%%
%%-----------------------------------------------------------------------
%% RCS: $Id: igscanner.xrl,v 1.4 1996/04/09 13:00:01 tobbe Exp $

Definitions.

%%---------------------------------------------------------
%% Some common definitions
%%---------------------------------------------------------
C_CHAR		"[^\n'\\]".
S_CHAR		"[^\n\"\\]".
LETTER		"[a-zA-Z]".
LETTER_UC_DIGIT	"[a-z_A-Z0-9]".
DIGIT		"[0-9]".
NONZERO_DIGIT	"[1-9]".
OCTAL_DIGIT	"[0-7]".
HEX_DIGIT	"[0-9A-Fa-f]".
HEX_PREFIX	"0[xX]".
ZERO		"0".
UC		"_".
INTEGER_SUFFIX	"[lLuU]".
FLOATING_SUFFIX	"[fFlL]".
EXPONENT	"(e[-+]*|E[-+]*){DIGIT}+".
COMMENT		"/\\*/*([^*/]|[^*]/|\\*[^/])*\\**\\*/".
WHITE		"[\000- ]".
HASH_DEFINE	"IG_define".
HASH_IFDEF	"IG_ifdef".
HASH_IFNDEF	"IG_ifndef".

IG_CMD		"/\\*IG\\*/".
IG_VAR		"/\\*IG_var\\*/".

%%---------------------------------------------------------
%% Escape characters
%%---------------------------------------------------------
CHAR_ESCAPE	"[ntbrfva'\"?\\]".
OCTAL_ESCAPE	"{OCTAL_DIGIT}{OCTAL_DIGIT}?{OCTAL_DIGIT}?".
HEX_ESCAPE	"x{HEX_DIGIT}+".
ESCAPE_CODE	"({CHAR_ESCAPE}|{OCTAL_ESCAPE}|{HEX_ESCAPE})".
ESCAPE_CHAR	"\\\\{ESCAPE_CODE}".

%%---------------------------------------------------------
%% For string handling
%%---------------------------------------------------------
S_CHAR_SEQ	"({S_CHAR}|{ESCAPE_CHAR})".

Rules.

%%---------------------------------------------------------
%% Operators
%%---------------------------------------------------------

"[][]"		mkpair(YYline,list_to_atom(YYtext)).
"[$_,&*+~!<>^=:?/%.]"
		mkpair(YYline,list_to_atom(YYtext)).
"[)(]"
		Token = list_to_atom(YYtext),
		mkpair(YYline,scope_analysis(Token)).

"-"		mkpair(YYline,list_to_atom(YYtext)).

"\\->"		mkpair(YYline,arrow).
"\\+\\+"	mkpair(YYline,icr).
"\\-\\-"	mkpair(YYline,decr).
"<<"		mkpair(YYline,ls).
">>"		mkpair(YYline,rs).
"<="		mkpair(YYline,le).
">="		mkpair(YYline,ge).
"=="		mkpair(YYline,eq).
"!="		mkpair(YYline,ne).
"&&"		mkpair(YYline,andand).
"\\|\\|"	mkpair(YYline,oror).
"\\.\\.\\."	mkpair(YYline,ellipsis).
"\\*="		mkpair(YYline,multassign).
"/="		mkpair(YYline,divassign).
"\\%="		mkpair(YYline,modassign).
"\\+="		mkpair(YYline,plusassign).
"\\-="		mkpair(YYline,minusassign).
"&="		mkpair(YYline,andassign).
"\\^="		mkpair(YYline,erassign).
"\\|="		mkpair(YYline,orassign).
"[\\}\\{;]"	
	Token = list_to_atom(YYtext),
	typedef_analysis(Token),
	mkpair(YYline,scope_analysis(Token)).

%%---------------------------------------------------------
%% Identifier or Reserved word 
%%
%%  Allowing ordinary identifiers to be type specifiers
%%  makes the C grammar context sensitive, and hence not
%%  LALR(1). Therefore we must do some 'typedef' analysis
%%  during the lexical analysis.
%%---------------------------------------------------------

"({LETTER}|{UC}){LETTER_UC_DIGIT}*"	
	Token = list_to_atom(YYtext),
	case reserved_word(Token) of
		true  -> mkpair(YYline,typedef_analysis(Token));
		false -> identifier_or_typedefname(YYline,Token)
	end.

%%---------------------------------------------------------
%% Integer constant
%%---------------------------------------------------------

%% Decimal constant
"{NONZERO_DIGIT}{DIGIT}*{INTEGER_SUFFIX}?"	
	{Suff,Stext} = strip_suffix(YYrevtext,"lLuU"),
	Const = add_suffix(list_to_integer(Stext),Suff),
	{integerconstant,YYline,Const}.

%% Octal constant
"{ZERO}{OCTAL_DIGIT}*{INTEGER_SUFFIX}?"	
	case strip_suffix(YYrevtext,"lLuU") of
		{Suff,[$0]} -> L = [$0];
		{Suff,Stext} -> 
			%% Strip off the leading zeroes
			L = string:strip(Stext,left,$0)
	end,
	Const = add_suffix(L,Suff),
	{octalconstant,YYline,Const}.

%% Hexadecimal constant
"{HEX_PREFIX}{HEX_DIGIT}+{INTEGER_SUFFIX}?"
	{Suff,Stext} = strip_suffix(YYrevtext,"lLuU"),
	%% Strip off the leading 0x
	L = tl(tl(Stext)),
	Const = add_suffix(L,Suff),
	{hexconstant,YYline,Const}.

%%---------------------------------------------------------
%% Floating-point constant
%%---------------------------------------------------------

"{DIGIT}+{EXPONENT}{FLOATING_SUFFIX}?"
	{Suff,Stext} = strip_suffix(YYrevtext,"fFlL"),
	Const = add_suffix(list_to_float(Stext),Suff),
	{floatingconstant,YYline,Const}.

"({DIGIT}+\\.{DIGIT}*|\\.{DIGIT}+){EXPONENT}?{FLOATING_SUFFIX}?"
	{Suff,Stext} = strip_suffix(YYrevtext,"fFlL"),
	Const = add_suffix(list_to_float(Stext),Suff),
	{floatingconstant,YYline,Const}.

%%---------------------------------------------------------
%% Character constant
%%
%%  We don't allow character constants such as 'ABCD' since
%%  this leads to portability problems.
%%
%%  ANSI C wide character constants, designated by the prefix 
%%  letter L ,have the type wchar_h defined in stddef.h .
%%---------------------------------------------------------

"[L]?'({C_CHAR}|{ESCAPE_CHAR})'"
	{Px,Stext} = strip_prefix(YYtext),
	Ch = mk_char_const(Stext),
	{characterconstant,YYline,add_suffix(Ch,Px)}.

%%---------------------------------------------------------
%% String constant
%%---------------------------------------------------------

"[L]?\"{S_CHAR_SEQ}*\""
	{Px,Stext} = strip_prefix(YYtext),
	String = string:strip(Stext,both,$"),
	{stringconstant,YYline,add_suffix(String,Px)}.

%%---------------------------------------------------------
%% Some special treatment of '#define' stuff
%%---------------------------------------------------------
%"{HASH_DEFINE}.*"	{hdef,YYline,YYtext}. 	
%"{HASH_IFDEF}.*"	{ifdef,YYline,YYtext}. 	
%"{HASH_IFNDEF}.*"	{ifndef,YYline,YYtext}. 	

"{IG_CMD}.#"		{ig_preproc, YYline}.

%% Q.A.D. should really change yecc spec instead /PL
"{IG_VAR}"		{'IG_var', YYline}. 

%%---------------------------------------------------------
%% Ignore stuff
%%---------------------------------------------------------
"{WHITE}". 	%% whitespace
"#.*".	   	%% Ignore Macro stuff for now
"{COMMENT}".	%% Ignore Comments

Erlang code.

-export([reserved_word/1]).

%%---------------------------------------------------------
%% Scope analysis
%%
%% The idea here is to make it possible to parse statement
%% for statement. So we must know when a ';' token is the
%% end-of-statement delimiter.
%%---------------------------------------------------------
scope_analysis(Token) ->
	case get(scope) of
		I when integer(I) ->
			scope_analysis(Token,I);
		_ ->
			put(scope,0),
			scope_analysis(Token,0)
	end.

scope_analysis(_,I) when I<0 -> {error,out_of_scope};
scope_analysis(';',0) -> end_of_statement;
scope_analysis(';',_) -> ';';
scope_analysis('{',I) -> inc_scope(I),'{';
scope_analysis('(',I) -> inc_scope(I),'(';
scope_analysis('}',I) -> dec_scope(I),'}';
scope_analysis(')',I) -> dec_scope(I),')';
scope_analysis(Tok,I) -> Tok.

inc_scope(I) -> put(scope,I+1).
dec_scope(I) -> put(scope,I-1).

%%---------------------------------------------------------
%% Typedef analysis
%%
%% When we encounter a 'typedef' token we set a
%% counter to 1 meaning that we have entered the
%% typedef-scope. If we encounters '{' or '}'
%% when we are in typedef-scope we increases or
%% decreases that counter. The idea is that when
%% we encounters an 'identifier' and the counter is
%% exactly 1 , then it is a 'typedef-name' token.
%%---------------------------------------------------------
typedef_analysis(Token) -> 
	case get(typedef) of
		I when integer(I) ->
			typedef_analysis(Token,I);
		_ ->
			put(typedef,0),
			typedef_analysis(Token,0)
	end.

typedef_analysis('typedef',_) -> 
	put(typedef,1),
	'typedef';
typedef_analysis(';',1) -> 
	put(typedef,0),
	';';
typedef_analysis('{',I) when integer(I),I>0 -> 
	put(typedef,I+1),
	'{';
typedef_analysis('}',I) when integer(I),I>0 -> 
	put(typedef,I-1),
	'}';
typedef_analysis(Token,_) -> 
	Token.
%%---------------------------------------------------------
%% An identifier that is currently in scope as a 
%% typedef name is provided to the parser as a 'typedefname'
%%---------------------------------------------------------
identifier_or_typedefname(Line,Token) ->
	identifier_or_typedefname(Line,Token,get(typedef)).

identifier_or_typedefname(Line,Token,1) -> 
	insert_typedef_name(Token),
	{typedefname,Line,Token};
identifier_or_typedefname(Line,Token,_) -> 
	case get(typedef_names) of
		Names when list(Names) ->
			case lists:member(Token,Names) of
				true  -> {typedefname,Line,Token};
				false -> {identifier,Line,Token}
			end;
		_ -> {identifier,Line,Token}
	end.

insert_typedef_name(Token) ->
	case get(typedef_names) of
		Names when list(Names) ->
			put(typedef_names,insert_unique(Token,Names));
		_ -> 
			put(typedef_names,[Token])
	end.

insert_unique(H,[H|T]) -> [H|T];
insert_unique(X,[H|T]) -> 
	[H|insert_unique(X,T)];
insert_unique(X,[]) -> [X].

mk_char_const(Stext) ->
	case string:strip(Stext,both,$') of
		[$\\,$x|T] -> base_convert(T,16);
		[$\\,T] when T>57 -> escape_convert(T);
		[$\\|T] -> base_convert(T,8);
		[T] -> T
	end.

strip_prefix(List) ->
	case List of
		[$L|S] -> {wchar_t,S};
		S -> {none,S}
	end.

reserved_word('IG_define') -> true;
reserved_word('IG_ifdef')  -> true;
reserved_word('IG_ifndef') -> true;
reserved_word('ifdef')     -> true;
reserved_word('ifndef')    -> true;
reserved_word('define')    -> true;
reserved_word('endif')	   -> true;
reserved_word('IG_else')   -> true;
reserved_word('IG_endif')  -> true;
reserved_word('IG_fun')    -> true;
reserved_word('IG_var')    -> true;
reserved_word('IG_void')   -> true;
reserved_word('IG_string') -> true;
reserved_word('IG_binaryPtr') -> true;
reserved_word('IG_call')   -> true;
reserved_word('IG_cast')   -> true;
reserved_word('IG_extern') -> true;
reserved_word('auto')     -> true;
reserved_word('break')    -> true;
reserved_word('case')     -> true;
reserved_word('char')     -> true;
reserved_word('const')    -> true;
reserved_word('continue') -> true;
reserved_word('default')  -> true;
reserved_word('do')       -> true;
reserved_word('double')   -> true;
reserved_word('else')     -> true;
reserved_word('enum')     -> true;
reserved_word('extern')   -> true;
reserved_word('float')    -> true;
reserved_word('for')      -> true;
reserved_word('goto')     -> true;
reserved_word('if')       -> true;
reserved_word('int')      -> true;
reserved_word('long')     -> true;
reserved_word('register') -> true;
reserved_word('return')   -> true;
reserved_word('short')    -> true;
reserved_word('signed')   -> true;
reserved_word('sizeof')   -> true;
reserved_word('static')   -> true;
reserved_word('struct')   -> true;
reserved_word('switch')   -> true;
reserved_word('typedef')  -> true;
reserved_word('union')    -> true;
reserved_word('unsigned') -> true;
reserved_word('void')     -> true;
reserved_word('volatile') -> true;
reserved_word('while')    -> true;
reserved_word(_)          -> false.

mkpair(Line,end_of_statement) -> {end_of_statement,Line};
mkpair(Line,{error,Reason}) -> {error,Reason,Line};
%%mkpair(Line,'IG_else') -> {else,Line};  %% Q.A.D
%%mkpair(Line,'IG_endif') -> {endif,Line}; %% Q.A.D
mkpair(Line,Sym) -> {Sym,Line}.

base_convert(L,Base) ->
    base(L, Base, 0).

base([C|Cs], Base, SoFar) when C >= $0, C =< $9, C < Base + $0 ->
    Next = SoFar * Base + (C - $0),
    base(Cs, Base, Next);
base([C|Cs], Base, SoFar) when C >= $a, C =< $f, C < Base + $a - 10 ->
    Next = SoFar * Base + (C - $a + 10),
    base(Cs, Base, Next);
base([C|Cs], Base, SoFar) when C >= $A, C =< $F, C < Base + $A - 10 ->
    Next = SoFar * Base + (C - $A + 10),
    base(Cs, Base, Next);
base([], _, N) ->
    N.

%% Remove a possible suffix and tag the result
strip_suffix([$u|Text],Sx) -> 
	case lists:member($u,Sx) of
		true  -> {unsigned,lists:reverse(Text)};
		false -> {none,lists:reverse([$u|Text])}
	end;
strip_suffix([$U|Text],Sx) -> 
	case lists:member($U,Sx) of
		true  -> {unsigned,lists:reverse(Text)};
		false -> {none,lists:reverse([$U|Text])}
	end;
strip_suffix([$l|Text],Sx) -> 
	case lists:member($l,Sx) of
		true  -> {long,lists:reverse(Text)};
		false -> {none,lists:reverse([$l|Text])}
	end;
strip_suffix([$L|Text],Sx) -> 
	case lists:member($L,Sx) of
		true  -> {long,lists:reverse(Text)};
		false -> {none,lists:reverse([$L|Text])}
	end;
strip_suffix([$f|Text],Sx) -> 
	case lists:member($f,Sx) of
		true  -> {float,lists:reverse(Text)};
		false -> {none,lists:reverse([$f|Text])}
	end;
strip_suffix([$F|Text],Sx) -> 
	case lists:member($F,Sx) of
		true  -> {float,lists:reverse(Text)};
		false -> {none,lists:reverse([$F|Text])}
	end;
strip_suffix(Text,_)  -> {none,lists:reverse(Text)}.

add_suffix(I,unsigned) -> {unsigned,I};
add_suffix(I,long)     -> {long,I};
add_suffix(I,float)    -> {float,I};
add_suffix(I,wchar_t)  -> {wchar_t,I};
add_suffix(I,_)        -> I.

escape_convert($n) -> $\n;	%\n = LF
escape_convert($r) -> $\r;	%\r = CR
escape_convert($t) -> $\t;	%\t = TAB
escape_convert($v) -> $\v;	%\v = VT
escape_convert($b) -> $\b;	%\b = BS
escape_convert($f) -> $\f;	%\f = FF
escape_convert($a) -> 7;	%\a = BEL = 007
escape_convert(C)  -> C.


