%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Interface functions.

%% string(String)
%% string(String, StartLine)
%%  Scan all the tokens in String.

string(String) ->
    string(String, 1, []).

string(String, Line) ->
    string(String, Line, []).

string([], Line, Ts) ->
    yyrev(Ts, []);
string(Ics0, Line0, Ts) ->
    case yystate(yystate(), Ics0, Line0, [], reject) of
	{accept,Ics,Line,Tcs,A} ->
	    string(Ics, Line, yyaction(A, Tcs, Line0, Ts));
	{{AIcs,ATcs,A},CurS,Ics,Line,Tcs} ->
	    string(AIcs, Line, yyaction(A, ATcs, Line0, Ts));
	{reject,CurS,[],Line,[]} ->		%Reached the end
	    yyrev(Ts, []);
	{reject,CurS,[],Line,Tcs} ->
	    yyrev(Ts, [{error,yyrev(Tcs)}]);
	{reject,CurS,[C|Ics],Line,[]} ->	%Bad character
	    string(Ics, Line, [{error,[C]}|Ts]);
	{reject,CurS,[C|Ics],Line,Tcs} ->	%Bad sequence
	    string([C|Ics], Line, [{error,yyrev(Tcs)}|Ts])
    end.

%% token(Continuation, Chars)
%% token(Continuation, Chars, StartLine)
%%  Reentrant function to scan one token. Intended to be used within
%%  io system. If we could not detect the end of a partial token then
%%  request more characters and try again.

token(Cont, Chars) ->
    token(Cont, Chars, 1).

token([], Chars, Line) ->		%First call
    token(yystate(), Chars, Line, [], reject);
token({State,Line,Tcs,Accept}, Chars, _) ->	%Subsequent calls
    token(State, Chars, Line, Tcs, Accept).

token(State, Ics0, Line0, Tcs0, Accept0) ->
    case yystate(State, Ics0, Line0, Tcs0, Accept0) of
	{accept,eof,Line,Tcs,A} ->
	    token_cont(yyaction(A, Tcs, Line0, []), [], Line);
	{accept,Ics,Line,Tcs,A} ->
	    token_cont(yyaction(A, Tcs, Line0, []), Ics, Line);
	{{AIcs,ATcs,A},CurS,[C|Ics],Line,Tcs} ->	%Could detect end
	    token_cont(yyaction(A, ATcs, Line, []), AIcs, Line);
	{{AIcs,ATcs,A},CurS,[],Line,Tcs} ->		%Could not detect end
	    {more,{CurS,Line,Tcs,{AIcs,ATcs,A}}};
	{{AIcs,ATcs,A},CurS,eof,Line,Tcs} ->		%Could detect end
	    token_cont(yyaction(A, ATcs, Line0, []), AIcs, Line);
	{reject,CurS,[C|Ics],Line,[]} ->
	    {done,{error,[C]},Ics};
	{reject,CurS,[C|Ics],Line,Tcs} ->
	    {done,{error,yyrev(Tcs)},[C|Ics]};
	{reject,CurS,[],Line,Tcs} ->			%Could not detect end
	    {more,{CurS,Line,Tcs,reject}};
	{reject,CurS,eof,Line,[]} ->
	    {done,eof,[]};
	{reject,CurS,eof,Line,Tcs} ->
	    {done,{error,yyrev(Tcs)},[]}
    end.

token_cont([Tok|_], Rest, Line) ->
    {done,{ok,Tok,Line},Rest};
token_cont([], Rest, Line) ->
    token(yystate(), Rest, Line, [], reject).

%% tokens(Continuation, Chars)
%% tokens(Continuation, Chars, StartLine)
%%  Reentrant function to scan tokens until a '$end' token. Intended to
%%  be used within io system.

tokens(Cont, Chars) ->
    tokens(Cont, Chars, 1).

tokens([], Chars, Line) ->				%First call
    tokens(yystate(), Chars, Line, [], reject, []);
tokens({State,Line,Tcs,Accept,Ts}, Chars, _) ->	%Subsequent calls
    tokens(State, Chars, Line, Tcs, Accept, Ts).

tokens(State, Ics0, Line0, Tcs0, Accept0, Ts) ->
    case yystate(State, Ics0, Line0, Tcs0, Accept0) of
	{accept,eof,Line,Tcs,A} ->
	    tokens_cont(yyaction(A, Tcs, Line0, Ts), [], Line);
	{accept,Ics,Line,Tcs,A} ->
	    tokens_cont(yyaction(A, Tcs, Line0, Ts), Ics, Line);
	{{AIcs,ATcs,A},CurS,[C|Ics],Line,Tcs} ->	%Could detect end
	    tokens_cont(yyaction(A, ATcs, Line0, Ts), AIcs, Line);
	{{AIcs,ATcs,A},CurS,[],Line,Tcs} ->		%Could not detect end
	    {more,{CurS,Line,Tcs,{AIcs,ATcs,A},Ts}};
	{{AIcs,ATcs,A},CurS,eof,Line,Tcs} ->		%Could detect end
	    tokens_cont(yyaction(A, ATcs,Line0, Ts), AIcs, Line);
	{reject,CurS,[C|Ics],Line,[]} ->
	    {done,{error,[C]},Ics};
	{reject,CurS,[C|Ics],Line,Tcs} ->
	    {done,{error,yyrev(Tcs)},[C|Ics]};
	{reject,CurS,[],Line,Tcs} ->			%Could not detect end
	    {more,{CurS,Line,Tcs,reject,Ts}};
	{reject,CurS,eof,Line,[]} ->
	    {done,eof,[]};
	{reject,CurS,eof,Line,Tcs} ->
	    {done,{error,yyrev(Tcs)},[]}
    end.

tokens_cont(['$end'|Toks], Rest, Line) ->
    {done,{ok,yyrev(Toks, ['$end']),Line},Rest};
tokens_cont([{'$end',L}|Toks], Rest, Line) ->
    {done,{ok,yyrev(Toks, [{'$end',L}]),Line},Rest};
tokens_cont(Toks, Rest, Line) ->
    tokens(yystate(), Rest, Line, [], reject, Toks).

%% Local utility functions.

yyrev(L) ->
    yyrev(L, []).

yyrev([H|T], Stack) ->
    yyrev(T, [H|Stack]);
yyrev([], Stack) -> Stack.
