/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.0, (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.erlang.org/EPL1_0.txt
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Erlang-4.7.3, December, 1998.
 * 
 * The Initial Developer of the Original Code is Ericsson Telecom
 * AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
 * Telecom AB. All Rights Reserved.
 * 
 * Contributor(s): ______________________________________.''
 */
/**
 *  $Id: memsup.c,v 1.1 1996/05/09 12:45:41 mbj Exp $
 *
 *  File:     memsup.c
 *  Purpose:  Portprogram for supervision of memory usage.
 *  Author:   peter@erlang.ericsson.se
 *
 *  Synopsis: memsup
 *
 *  PURPOSE OF THIS PROGRAM
 *
 *  This program supervises the memory status of the entire system, and
 *  sends status reports upon request from the Erlang system
 *
 *  SPAWNING FROM ERLANG
 *
 *  This program is started from Erlang as follows,
 *
 *      Port = open_port({spawn, 'memsup'}, [{packet,1}]) for UNIX and VxWorks
 *
 *  Erlang sends one of the request condes defined in memsup.h and this
 *  program answers in one of two ways:
 *  * If the request is for simple memory data (which is used periodically
 *    for monitoring) the answer is simply sent in two packets, where 
 *    all numbers are divided by 256 (shifted right 8 steps) to avoid
 *    problems with HUGE memories.
 *  * If the request is for the system specific data, the answer is delivered
 *    in two packets per value, first a tag value, then the actual
 *    value. The values are delivered "as is", this interface is
 *    mainly for VxWorks.
 *
 *  FAKING
 *
 *  When we don't know how to get the memory usage from the Operating
 *  System we fake the report. The total physical memory is always
 *  reported to be 256MB, and the used fraction to be 128MB.
 *
 *  STANDARD INPUT, OUTPUT AND ERROR
 *
 *  This program communicates with Erlang through the standard
 *  input and output file descriptors (0 and 1). These descriptors
 *  (and the standard error descriptor 2) must NOT be closed
 *  explicitely by this program at termination (in UNIX it is
 *  taken care of by the operating system itself; in VxWorks
 *  it is taken care of by the spawn driver part of the Emulator).
 *
 *  END OF FILE
 *
 *  If a read from a file descriptor returns zero (0), it means
 *  that there is no process at the other end of the connection
 *  having the connection open for writing (end-of-file).
 *
 *  COMPILING
 *
 *  When the target is VxWorks the identifier VXWORKS must be defined for
 *  the preprocessor (usually by a -D option).
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>

#ifdef STDC_HEADERS
#include <stddef.h>
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef VXWORKS
#include <vxWorks.h>
#include <ioLib.h>
#include <memLib.h>
#endif

/* commands */
#include "memsup.h"

#define ERLIN_FD      0
#define ERLOUT_FD     1


#ifdef VXWORKS
extern int erl_mem_info_get(MEM_PART_STATS *);

static MEM_PART_STATS latest;
static unsigned long latest_system_total; /* does not fit in the struct */
#endif

static char *program_name;

static void send(unsigned long value)
{
    char buf[20];
    int left, bytes, res;
    
    sprintf(buf+1, "%lu", value);
    bytes = (char) strlen(buf+1); 
    buf[0] = bytes;		
    left = ++bytes;

    while (left > 0) {
	res = write(ERLOUT_FD, buf+bytes-left, left);
	if (res <= 0) {
	    perror("Error writing to pipe");
	    exit(1);
	}
	left -= res;
    }
}

static void send_tag(int value)
{
    unsigned char buf[2];
    int res,left;

    buf[0] = 1U;
    buf[1] = (unsigned char) value;
    left = 2;

    while(left > 0) {
	if((res = write(ERLOUT_FD, buf+left-2,left)) <= 0){
	    perror("Error writing to pipe");
	    exit(1);
	} else {
	    left -= res;
	}
    }
}

#ifdef VXWORKS
static void load_statistics(void)
{
    if (memPartInfoGet(memSysPartId,&latest) != OK)
	memset(&latest,0,sizeof(latest));
    
    latest_system_total = latest.numBytesFree + latest.numBytesAlloc;
    
    erl_mem_info_get(&latest); /* if it fails, latest is untouched */
}
#endif


static void 
get_basic_mem(unsigned long *tot, unsigned long *used, int shift)
{
#if defined(VXWORKS)
    load_statistics();
    
    *tot = (latest.numBytesFree + latest.numBytesAlloc) >> shift;
    *used = latest.numBytesAlloc >> shift;
    
#elif defined(HAVE_SYSCONF) && defined(_SC_AVPHYS_PAGES)
    unsigned long avPhys, phys, pgSz;
    
    phys = sysconf(_SC_PHYS_PAGES);
    avPhys = sysconf(_SC_AVPHYS_PAGES);
    pgSz = sysconf(_SC_PAGESIZE) >> shift; /* shift pgsz before multipl. */
    *used = (phys - avPhys) * pgSz;
    *tot = phys * pgSz;
    
#else
    *used = (1<<27) >> shift;	/* Fake! 128 MB used */
    *tot = (1<<28) >> shift;	/* Fake! 256 MB total */
#endif
}    

/*
** The simple memory showing 
** sends data divided by 256 to avoid integer overflow on extreme
** large memory systems.
*/
static void simple_show_mem(void)
{
    unsigned long tot, used;

    get_basic_mem(&tot, &used, 8);
    
    send(used);
    send(tot);
}

/*
** The system memory showing sends data as expected (in bytes)
** and also tags values as output may be different on different 
** OS:es. Actually I expect this protocol to be really useful
** only on non paged systems like VxWorks, OSE/Delta etc.
** The possibilities of 4GB+ memories on these
** systems are not counted for...
*/
static void 
extended_show_mem(void){
    unsigned long tot, used;

     /* We don't care about extremely large memories here, we report
      * to erlang as bytes
      */
    get_basic_mem(&tot, &used, 0);
    
    send_tag(TOTAL_MEMORY);
    send(tot);
    send_tag(FREE_MEMORY);
    send(tot - used);
    send_tag(SYSTEM_TOTAL_MEMORY);
    
#ifdef VXWORKS
    send(latest_system_total);
    send_tag(LARGEST_FREE);
    send(latest.maxBlockSizeFree);
    send_tag(NUMBER_OF_FREE);
    send(latest.numBlocksFree);
#else
    send(tot);
#endif
    send_tag(SYSTEM_MEM_SHOW_END);
}    


static void message_loop(int erlin_fd)
{
    char cmdLen, cmd;
    
    while (1) {
	/*
	 * Wait for command from Erlang, one byte length and one
	 * byte commands.
	 */
	if (read(erlin_fd, &cmdLen, 1) != 1) {
	    fprintf(stderr, "%s: Error reading from Erlang.\n", program_name);
	    return;
	}

	if (cmdLen == 1) {	/* Should be! */
	    switch (read(erlin_fd, &cmd, 1)) {
	    case 1:
		switch (cmd){
		case MEM_SHOW:
		    simple_show_mem();
		    break;
		case SYSTEM_MEM_SHOW:
		    extended_show_mem();
		    break;
		default:	/* ignore all other messages */
		    break;
		}
		break;

	    case 0:
		fprintf(stderr, "%s: Erlang has closed.\n", program_name);
		return;

	    default:
		fprintf(stderr, "%s: Error reading from Erlang.\n",
			program_name);
		return;
	    } /* switch() */

	} else { /* cmdLen != 1 */
	    fprintf(stderr, "%s: Invalid command length (%d) received.",
		    program_name, cmdLen);
	    return;
	}
    }
}


#ifdef VXWORKS
#define MAIN memsup
#else
#define MAIN main
#endif

int MAIN(int argc, char **argv)
{
  program_name = argv[0];
  
  message_loop(ERLIN_FD);
  
  return 0;
}
